// This file is part of Lanradio.
//
// Lanradio is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Lanradio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Lanradio.  If not, see <http://www.gnu.org/licenses/>.
//
// Authors:
// Florian Klingler <klingler@ccs-labs.org>

#include "veins/modules/lanradio/LanradioCommunication.h"

#define MYDEBUG EV

using Veins::LanradioCommunication;

Define_Module(Veins::LanradioCommunication);

LanradioCommunication::~LanradioCommunication()
{
}

//  recv zmq string
static std::string zmq_recv(zmq::socket_t &sock) {

    zmq::message_t msg;
    sock.recv(&msg);

    return std::string(static_cast<char *>(msg.data()), msg.size());
}

//  send zmq string
static bool zmq_send(zmq::socket_t &sock, const std::string &str) {

    zmq::message_t msg(str.size());
    memcpy(msg.data(), str.data(), str.size());

    bool ret = sock.send(msg);
    return ret;
}

//  send zmq string, sendmore
static bool zmq_sndmore(zmq::socket_t &sock, const std::string &str) {

    zmq::message_t msg(str.size());
    memcpy(msg.data(), str.data(), str.size());

    bool ret = sock.send(msg, ZMQ_SNDMORE);
    return ret;
}

void thread_receive_messages(bool* running, LanradioCommunication* communication_module, bool lara_debug) {

    std::string messages_envelope = "";

    std::string messages_zmq_hostname = communication_module->getHostname();
    int messages_zmq_port = communication_module->getBasePort()+2;
    zmq::context_t* messages_zmq_context = new zmq::context_t(1);
    zmq::socket_t* messages_zmq_socket= new zmq::socket_t(*messages_zmq_context, ZMQ_SUB);
    messages_zmq_socket->setsockopt(ZMQ_SNDTIMEO, 1000);
    messages_zmq_socket->setsockopt(ZMQ_RCVTIMEO, 1000);

    messages_zmq_socket->connect(("tcp://" + messages_zmq_hostname + ":" + std::to_string(messages_zmq_port)).c_str());
    if (messages_envelope == "") {
        messages_zmq_socket->setsockopt(ZMQ_SUBSCRIBE, "", 0);//subscribe to all messages
    } else {
        messages_zmq_socket->setsockopt(ZMQ_SUBSCRIBE, messages_envelope.c_str(), 1);
    }


    while(*running) {

        std::string envelope = zmq_recv(*messages_zmq_socket);
        std::string message = zmq_recv(*messages_zmq_socket);


        if(envelope=="WLAN") {
            if(lara_debug) std::cerr << "received from hw: " << "\n\t" << "envelope: " << envelope << "\n\t" << "msg-length: " << message.length() << std::endl;
            wlanPackage::WLAN* wlan = new wlanPackage::WLAN();
            bool successful = wlan->ParseFromString(message);

            if(successful) {

                std::string byteMessage;
                byteMessage = wlan->payload();

                LanradioCommunication::wlan_msg m;
                m.src_mac = wlan->srcmac();
                m.dst_mac = wlan->dstmac();
                m.payload = byteMessage;

                if(wlan->ethertype() == ETHERTYPE_CAR) communication_module->insert_wlan_msg_buffer(m);
            }
        }
    }

    //messages_zmq_context->close();
    messages_zmq_socket->close();
    //delete messages_zmq_context;
    delete messages_zmq_socket;
}

void LanradioCommunication::insert_wlan_msg_buffer(wlan_msg m) {
    //Enter_Method_Silent();

    mutex_message_buffer.lock();
    if(lara_debug) std::cerr << "inserted in wlan_msg_buffer" << std::endl;

    wlan_message_buffer->insert(std::make_pair(simTime(), m));
    mutex_message_buffer.unlock();

}

void LanradioCommunication::handleMessage(cMessage *msg) {

    switch (msg->getKind()) {
    case Lanradio_CHECK_MSG_BUFFER: {
        process_message_buffer();
        scheduleAt(simTime()+0.001, check_message_buffer);
        break;
    }
    default: {
        if (msg) {
            if(lara_debug) std::cerr << "APP: Error: Got Self Message of unknown kind! Name: " << msg->getName() << endl;
            delete msg;
        }
        break;
    }
    }
}

void LanradioCommunication::process_message_buffer() {

    mutex_message_buffer.lock();

    for(std::map<simtime_t, wlan_msg>::iterator it_wlan=wlan_message_buffer->begin(); it_wlan!=wlan_message_buffer->end();) {

        received_WLAN(it_wlan->second);
        wlan_message_buffer->erase(it_wlan++);
    }

    mutex_message_buffer.unlock();
}


void LanradioCommunication::initialize(int stage)
{
    /*
     * setup zmq socket for configuring Lanradio
     */
    if(stage == 0) {

        use_lanradio = par("use_lanradio").boolValue();
        baseport = par("baseport").longValue();
        hostname = par("hostname").stringValue();
        wlan_interface = par("wlan_interface").stringValue();
        bitrates = par("bitrates").stringValue();
        wlan_ip = par("wlan_ip").stringValue();
        freq = par("freq").stringValue();
        bandwidth = par("bandwidth").stringValue();
        lara_debug = par("lara_debug").boolValue();


        if(use_lanradio) {
            radioconfig_zmq_hostname = hostname;
            radioconfig_zmq_port = baseport;

            radioconfig_zmq_context = new zmq::context_t(1);
            radioconfig_zmq_socket = new zmq::socket_t(*radioconfig_zmq_context, ZMQ_REQ);
            radioconfig_zmq_socket->setsockopt(ZMQ_SNDTIMEO, 1000);
            radioconfig_zmq_socket->setsockopt(ZMQ_RCVTIMEO, 1000);
            radioconfig_zmq_socket->connect(("tcp://" + radioconfig_zmq_hostname + ":" + std::to_string(radioconfig_zmq_port)).c_str());

            activate_lanradio();

            wlan_message_buffer = new std::map<simtime_t, wlan_msg>();

            if(lara_debug) std::cerr << "starting receive thread" << std::endl;
            receive_thread_running = new bool();
            *receive_thread_running = true;
            receive_thread = new std::thread(thread_receive_messages, receive_thread_running, this, lara_debug);

            check_message_buffer = new cMessage("check msg buffer", Lanradio_CHECK_MSG_BUFFER);
            scheduleAt(simTime(), check_message_buffer);
        }
    }

    if(use_lanradio) {

        std::string portOut = std::to_string(baseport+1);

        mContext = new zmq::context_t(1);
        mPublisher = new zmq::socket_t(*mContext, ZMQ_PUB);
        mPublisher->bind(("tcp://*:" + portOut).c_str());

        mPublisher->setsockopt(ZMQ_SNDTIMEO, 1000);
        mPublisher->setsockopt(ZMQ_RCVTIMEO, 1000);
    }
}

void LanradioCommunication::activate_radio() {

    radio_send_command("iw reg set 00");
    radio_send_command("ip link set "+wlan_interface+" down");
    radio_send_command("iw dev "+wlan_interface+" set type ocb");
    radio_send_command("ip link set "+wlan_interface+" up");
    radio_send_command("iw dev "+wlan_interface+" ocb join "+freq+" "+bandwidth);
    radio_send_command("ifconfig "+wlan_interface+" "+wlan_ip);
}

void LanradioCommunication::activate_lanradio() {

    radio_send_command("iw reg set DE");
    radio_send_command("ip link set "+wlan_interface+" down");
    radio_send_command("iw dev "+wlan_interface+" set type ocb");
    radio_send_command("ip link set "+wlan_interface+" up");
    radio_send_command("iw dev "+wlan_interface+" ocb join "+freq+" "+bandwidth);
    radio_send_command("ifconfig "+wlan_interface+" "+wlan_ip);
    radio_send_command("iw dev "+wlan_interface+" set txpower fixed "+txpower);
    radio_send_command("iw dev "+wlan_interface+" set bitrates "+bitrates);
}

void LanradioCommunication::deactivate_radio() {

    radio_send_command("iw dev "+wlan_interface+" ocb leave");
    radio_send_command("ip link set "+wlan_interface+" down");
    radio_send_command("ifconfig "+wlan_interface+" down");
}

std::string LanradioCommunication::radio_send_command(const std::string cmd) {

    std::string repl = "";
    try {
        bool sucessful = zmq_send(*radioconfig_zmq_socket, cmd);
        if(!sucessful) {
            error("connection to Lanradio was unsuccessful");
        }
        std::cout << "radioconfig\n\tsent: " << cmd << std::endl;
        repl = zmq_recv(*radioconfig_zmq_socket);

        std::cout << "\trcvd: " << repl << std::endl;
    } catch (zmq::error_t) {
        error("connection to Lanradio was unsuccessful");
    }

    return repl;
}

void LanradioCommunication::finish()
{
    *receive_thread_running = false;
    receive_thread->join();
    deactivate_radio();

    if(use_lanradio) {
        radioconfig_zmq_socket->close();
        radioconfig_zmq_context->close();
        delete radioconfig_zmq_socket;
        delete radioconfig_zmq_context;


        mPublisher->close();
        mContext->close();
        delete mPublisher;
        delete mContext;
    }
}

void LanradioCommunication::send_WLAN(void* data, t_wlan_type type, std::string srcmac, std::string dstmac) {

    if(type==wlan_CAM) {
        CAM_t* cam = static_cast<CAM_t*>(data);

        std::vector<uint8_t> encodedCam = MessageUtils::encodeMessage(&asn_DEF_CAM, cam);
        std::string strCam(encodedCam.begin(), encodedCam.end());

        std::string serializedData;
        wlanPackage::WLAN wlan;

        // create packet with geonet and btp header
        unsigned int geoHdrLen;
        uint8_t* geoHdr;

        fillGeoNetBTPheaderForCam(cam, strCam.size(), srcmac);
        geoHdrLen = sizeof(struct GeoNetworkAndBTPHeaderCAM);
        geoHdr = reinterpret_cast<uint8_t*>(&mGeoBtpHdrForCam);

        unsigned int packetsize = geoHdrLen + strCam.size();
        unsigned char packet[packetsize];
        unsigned char* payload = packet + geoHdrLen;
        unsigned char* geoNetHdr = packet;

        // Fill in geo networking and btp header
        memcpy(geoNetHdr, geoHdr, geoHdrLen);

        //copy payload to packet
        memcpy(payload,strCam.c_str(),strCam.size());

        // end create packet


        wlan.set_priority(wlanPackage::WLAN_Priority_BE);
        wlan.set_ethertype(ETHERTYPE_CAR);

        //std::string payload_string = std::string(packet);

        //std::cerr << "payload-length" << payload_string.size() << std::endl;
        wlan.set_payload(packet, packetsize);
        wlan.set_srcmac(srcmac);
        wlan.set_dstmac(dstmac);

        wlan.SerializeToString(&serializedData);

        zmq_sndmore(*mPublisher, "WLAN");
        zmq_send(*mPublisher, serializedData);

        asn_DEF_CAM.free_struct(&asn_DEF_CAM, cam, 0);

    } else if(type==wlan_SPAT) {
        SPAT_PDU_t* spat = static_cast<SPAT_PDU_t*>(data);

        std::vector<uint8_t> encodedSpat = MessageUtils::encodeMessage(&asn_DEF_SPAT_PDU, spat);
        std::string strSpat(encodedSpat.begin(), encodedSpat.end());

        std::string serializedData;
        wlanPackage::WLAN wlan;

        // create packet with geonet and btp header
        unsigned int geoHdrLen;
        uint8_t* geoHdr;

        fillGeoNetBTPheaderForSpat(strSpat.size(), srcmac);
        geoHdrLen = sizeof(struct GeoNetworkAndBTPHeaderSPAT);
        geoHdr = reinterpret_cast<uint8_t*>(&mGeoBtpHdrForSpat);

        unsigned int packetsize = geoHdrLen + strSpat.size();
        unsigned char packet[packetsize];
        unsigned char* payload = packet + geoHdrLen;
        unsigned char* geoNetHdr = packet;

        // Fill in geo networking and btp header
        memcpy(geoNetHdr, geoHdr, geoHdrLen);

        //copy payload to packet
        memcpy(payload,strSpat.c_str(),strSpat.size());

        // end create packet


        wlan.set_priority(wlanPackage::WLAN_Priority_BE);
        wlan.set_ethertype(ETHERTYPE_CAR);

        //std::string payload_string = std::string(packet);

        //std::cerr << "payload-length" << payload_string.size() << std::endl;
        wlan.set_payload(packet, packetsize);
        wlan.set_srcmac(srcmac);
        wlan.set_dstmac(dstmac);

        wlan.SerializeToString(&serializedData);

        zmq_sndmore(*mPublisher, "WLAN");
        zmq_send(*mPublisher, serializedData);

        asn_DEF_SPAT_PDU.free_struct(&asn_DEF_SPAT_PDU, spat, 0);
    }

}

void LanradioCommunication::fillGeoNetBTPheaderForCam(CAM_t* cam, int payloadLen, std::string srcmac) {
    // GeoNetwork Header
    mGeoBtpHdrForCam.mGeoNetHdr.basicHeader.versionAndNH = 1;
    mGeoBtpHdrForCam.mGeoNetHdr.basicHeader.reserved = 0;
    mGeoBtpHdrForCam.mGeoNetHdr.basicHeader.lifetime = 241;
    mGeoBtpHdrForCam.mGeoNetHdr.basicHeader.remainingHopLimit = 1;

    mGeoBtpHdrForCam.mGeoNetHdr.commonHeader.nhAndReserved = 32;
    mGeoBtpHdrForCam.mGeoNetHdr.commonHeader.htAndHst = 80;
    mGeoBtpHdrForCam.mGeoNetHdr.commonHeader.tc = 2;
    mGeoBtpHdrForCam.mGeoNetHdr.commonHeader.payload = htons(payloadLen + sizeof(struct BTPHeader));
    mGeoBtpHdrForCam.mGeoNetHdr.commonHeader.maxHop = 1;
    mGeoBtpHdrForCam.mGeoNetHdr.commonHeader.reserved = 0;
    mGeoBtpHdrForCam.mGeoNetHdr.tsb.spv.addr.assignmentTypeCountryCode = htons(38393);
    //mGeoBtpHdrForCam.mGeoNetHdr.tsb.spv.addr.llAddr = reinterpret_cast<uint8_t*>(ether_aton(mOwnMac.c_str()));
    memcpy(&mGeoBtpHdrForCam.mGeoNetHdr.tsb.spv.addr.llAddr, ether_aton(srcmac.c_str()), ETH_ALEN);
    mGeoBtpHdrForCam.mGeoNetHdr.tsb.spv.timestamp = htonl(2810450329);
    mGeoBtpHdrForCam.mGeoNetHdr.tsb.spv.latitude = htonl(cam->cam.camParameters.basicContainer.referencePosition.latitude);//htonl(424937722);
    mGeoBtpHdrForCam.mGeoNetHdr.tsb.spv.longitude = htonl(cam->cam.camParameters.basicContainer.referencePosition.longitude); //htonl(3460636913);
    mGeoBtpHdrForCam.mGeoNetHdr.tsb.spv.speed = htons(496);
    mGeoBtpHdrForCam.mGeoNetHdr.tsb.spv.heading = htons(1996);

    mGeoBtpHdrForCam.mGeoNetHdr.tsb.reserved = htonl(0);

    // BTP Header
    mGeoBtpHdrForCam.mBTPHdr.mDestinationPort = htons(2001);
    mGeoBtpHdrForCam.mBTPHdr.mSourcePort = htons(0);
    //uint8_t* temp = reinterpret_cast<uint8_t*>(&mGeoBtpHdrForCam);
    //dumpBuffer(temp, sizeof(mGeoBtpHdrForCam));
}

void LanradioCommunication::fillGeoNetBTPheaderForSpat(int payloadLen, std::string srcmac) {
    // GeoNetwork Header
    mGeoBtpHdrForSpat.mGeoNetHdr.basicHeader.versionAndNH = 1;
    mGeoBtpHdrForSpat.mGeoNetHdr.basicHeader.reserved = 0;
    mGeoBtpHdrForSpat.mGeoNetHdr.basicHeader.lifetime = 241;
    mGeoBtpHdrForSpat.mGeoNetHdr.basicHeader.remainingHopLimit = 1;

    mGeoBtpHdrForSpat.mGeoNetHdr.commonHeader.nhAndReserved = 32;
    mGeoBtpHdrForSpat.mGeoNetHdr.commonHeader.htAndHst = 80;
    mGeoBtpHdrForSpat.mGeoNetHdr.commonHeader.tc = 2;
    mGeoBtpHdrForSpat.mGeoNetHdr.commonHeader.payload = htons(payloadLen + sizeof(struct BTPHeader));
    mGeoBtpHdrForSpat.mGeoNetHdr.commonHeader.maxHop = 1;
    mGeoBtpHdrForSpat.mGeoNetHdr.commonHeader.reserved = 0;
    mGeoBtpHdrForSpat.mGeoNetHdr.tsb.spv.addr.assignmentTypeCountryCode = htons(38393);
    //mGeoBtpHdrForSpat.mGeoNetHdr.tsb.spv.addr.llAddr = reinterpret_cast<uint8_t*>(ether_aton(mOwnMac.c_str()));
    memcpy(&mGeoBtpHdrForSpat.mGeoNetHdr.tsb.spv.addr.llAddr, ether_aton(srcmac.c_str()), ETH_ALEN);
    mGeoBtpHdrForSpat.mGeoNetHdr.tsb.spv.timestamp = htonl(2810450329);
    mGeoBtpHdrForSpat.mGeoNetHdr.tsb.spv.latitude = htonl(424937722);
    mGeoBtpHdrForSpat.mGeoNetHdr.tsb.spv.longitude = htonl(3460636913);
    mGeoBtpHdrForSpat.mGeoNetHdr.tsb.spv.speed = htons(496);
    mGeoBtpHdrForSpat.mGeoNetHdr.tsb.spv.heading = htons(1996);

    mGeoBtpHdrForSpat.mGeoNetHdr.tsb.reserved = htonl(0);

    // BTP Header
    mGeoBtpHdrForSpat.mBTPHdr.mDestinationPort = htons(2004);
    mGeoBtpHdrForSpat.mBTPHdr.mSourcePort = htons(0);
    //uint8_t* temp = reinterpret_cast<uint8_t*>(&mGeoBtpHdrForSpat);
    //dumpBuffer(temp, sizeof(mGeoBtpHdrForSpat));
}

void LanradioCommunication::received_WLAN(wlan_msg m) {
    if(!use_lanradio) return;

    // Parse Geonet Header

    int mBytes = m.payload.size();
    int mLinkLayerLength = 0;//sizeof(struct ether_header);
    char* mPacket = (char*) m.payload.c_str();

    int geoNetPDULen = mBytes - mLinkLayerLength;

    //std::cerr << "mBytes = " << mBytes << ", mLinkLayerLength = " << mLinkLayerLength << ", geoNetPDULen=" << geoNetPDULen << std::endl;
    //for(int i = 0; i<50; i++) {
    //    printf(": %x \n", (unsigned char) *(mPacket+i));
    //}

    /* read geonet header */
    unsigned int geoHdrLen;
    uint8_t* geoHdr;

    //fillGeoNetBTPheaderForCam(strCam.size(), srcmac);
    geoHdrLen = sizeof(struct GeoNetworkAndBTPHeaderCAM);
    geoHdr = reinterpret_cast<uint8_t*>(&mGeoBtpHdrForCam);

    //unsigned char packet[geoNetPDULen];
    unsigned char* payload = mPacket + geoHdrLen;
    unsigned char* geoNetHdr = mPacket;

    // Fill in geo networking and btp header
    memcpy(geoHdr, geoNetHdr, geoHdrLen);

    GeoNetworkAndBTPHeaderCAM* myHeader = reinterpret_cast<GeoNetworkAndBTPHeaderCAM*>(geoHdr);
    //std::cerr << "got: " << myHeader->mBTPHdr.mDestinationPort << std::endl;

    //copy payload to packet
    //memcpy(payload,strCam.c_str(),strCam.size());
    /* end read geonet header */

    char* geoNetPDU = mPacket + mLinkLayerLength;
    if (/*geoNetPDU[5] == 80 && geoNetPDU[40,41,42,43]*/ntohs(myHeader->mBTPHdr.mDestinationPort) == 2001) {
        // CAM
        int camPDULen = geoNetPDULen - sizeof(struct GeoNetworkAndBTPHeaderCAM);
        char* camPDU = geoNetPDU + sizeof(struct GeoNetworkAndBTPHeaderCAM);
        std::string serializedAsnCam(camPDU, camPDULen);
        //ReceivedPacketInfo info;
        //info.mSenderMac = senderMac;
        //info.mType = dataPackage::DATA_Type_CAM;
        //info.was_outgoing_packet = (addr.sll_pkttype == PACKET_OUTGOING);

        CAM_t* cam = 0;
        int res = MessageUtils::decodeMessage(&asn_DEF_CAM, (void **)&cam, serializedAsnCam);
        if (res != 0) {
            //mLogger->logError("Failed to decode received CAM. Error code: " + to_string(res));
            if(lara_debug) std::cerr << "Failed to decode received CAM. Error code: " << std::to_string(res) << std::endl;
        }
        //asn_fprint(stdout, &asn_DEF_CAM, cam);

        if(lara_debug) std::cerr << "success: stationId=" << cam->header.stationID<< std::endl;

        CAMMessage* cammsg = new CAMMessage("CAM");

        cammsg->setWsmVersion(1);
        cammsg->setTimestamp(simTime());
        cammsg->setSenderAddress(-1);
        cammsg->setRecipientAddress(-1);
        cammsg->setBitLength(0);
        cammsg->setPsid(-1);
        cammsg->setChannelNumber(Channels::CCH);
        cammsg->addBitLength(100);
        cammsg->setUserPriority(2);
        cammsg->setSrcmac(m.src_mac.c_str());

        cammsg->setCam(*cam);

        intermediate_mac->handleLanradioMessage(cammsg);

        // to test loopback: echo back received packets.
        //std::cerr << "again sending CAM" << std::endl;
        //send_WLAN(cam, LanradioCommunication::wlan_CAM, m.src_mac, m.dst_mac);

    } else if (/*geoNetPDU[5] == 66*/ntohs(myHeader->mBTPHdr.mDestinationPort) == 2002) {
        // DENM
        error("we currently do not support DENMs");

    } else if(myHeader->mBTPHdr.mDestinationPort == htons(2004)) {
        if(lara_debug) std::cerr << "received a SPAT, no further processing" << std::endl;
    } else {
        // TODO: Possible cases?
        error("we currently do not support anything else than CAMs and SPATs");
    }
    // end parse Geonet Header
}

bool LanradioCommunication::registerRadio(IntermediateMacInterface* mac) {
    if(!use_lanradio) return true;

    intermediate_mac = mac;

    return true;
}
